# frozen_string_literal: true

require_relative 'node'

module Modelur
  # Earcut Library for polygon triangulation (with holes).
  # Ported from earcut-python (https://github.com/joshuaskelly/earcut-python),
  # which itself is ported javascript earcut (https://github.com/mapbox/earcut).
  module Earcut
    # Main earcut algorithm that returns triangle indicies.
    #
    # @param vertices is a flat [Array] of vertex coordinates like
    # [x0,y0, x1,y1, x2,y2, ...].
    # @param hole_indicies is an [Array] of hole indices if any (e.g. [5, 8] for a
    # 12-vertex input would mean one hole with vertices 5–7 and another with 8–11).
    # @param dimensions is the [Integer] of coordinates per vertex in the input
    # array (2 by default).
    #
    # @return [Array<Integer>] Each group of three vertex indices in the resulting array forms
    #  a triangle (e.g. [1, 0, 3, 3, 2, 1])
    #
    # @example
    #   triangles = Modelur::Earcut::earcut([10,0, 0,50, 60,60, 70,10], nil, 2)
    # 	#=> [1, 0, 3, 3, 2, 1]
    def self.earcut(vertices, hole_indicies = nil, dim = nil)
      dim = dim || 2
      has_holes = (!hole_indicies.nil? and hole_indicies.any?)

      if has_holes
        outer_length = hole_indicies[0] * dim
      else
        outer_length = vertices.length
      end

      outer_node = linked_list(vertices, 0, outer_length, dim, true)
      triangles = []
      return triangles unless outer_node

      min_x = nil
      min_y = nil
      size = nil

      if has_holes
        outer_node = eliminate_holes(vertices, hole_indicies, outer_node, dim)
      end

      # if the shape is not too simple, we'll use z-order curve hash later;
      # calculate polygon bbox
      if vertices.size > 80 * dim
        min_x = max_x = vertices[0]
        min_y = max_y = vertices[1]
        vertices.each_cons(dim) do |x, y, _z|
          if x < min_x
            min_x = x
          end
          if y < min_y
            min_y = y
          end
          if x > max_x
            max_x = x
          end
          if y > max_y
            max_y = y
          end
        end
        # minX, minY and size are later used to transform coords into integers
        # for z-order calculation
        size = [max_x - min_x, max_y - min_y].max
      end
      earcut_linked(outer_node, triangles, dim, min_x, min_y, size)
      triangles
    end

    # If your input is a multi-dimensional array (e.g. GeoJSON Polygon), you can
    # convert it to the format expected by Earcut with .flatten method.
    # Note that the first element in array represents outer loop and all
    # elements following it represent holes in the polygon.
    #
    # @param coordinates [Array] of GeoJSON 2D Polygon coordinates (eg.
    # [[[0, 0], [10, 0], [10,10], [0, 10]]])
    #
    # data = earcut.flatten(geojson.geometry.coordinates);
    #
    # @example
    #   coords = [[[0, 0], [10, 0], [10,10], [0, 10]]]
    #   data = Modelur::Earcut::flatten(coords)
    #   #=> {"vertices"=>[0, 0, 10, 0, 10, 10, 0, 10], "holes"=>[], "dimensions"=>2}
    #   # polygon with hole
    # 	coords = [[[0,0],[10,0],[10,10],[0,10]], [[2,2],[8,2],[8,8],[2,8]]]
    # 	data = Modelur::Earcut::flatten(coords)
    #		#=> {"vertices"=>[0, 0, 10, 0, 10, 10, 0, 10, 2, 2, 8, 2, 8, 8, 2, 8], "holes"=>[4], "dimensions"=>2}
    def self.flatten(coordinates)
      # puts "coordinates = #{coordinates}"
      dim = coordinates[0][0].length
      result = {"vertices" => [], "holes" => [], "dimensions" => dim}
      hole_index = 0
      coordinates.size.times.each { |i|
        coordinates[i].size.times.each { |j|
          dim.times.each { |d|
            result["vertices"].push(coordinates[i][j][d])
          }
        }
        if i > 0
          hole_index += (coordinates[i - 1]).size
          result["holes"].push(hole_index)
        end
      }
      result
    end

    private

    # Create a circular doubly linked _list from polygon points in the specified
    # winding order.
    #
    # @param data Add description
    # @param start_node Add description
    # @param end_node Add description
    # @param dim Add description
    # @param clockwise Add description
    #
    # @example
    #   Add example.
    def self.linked_list(data, start_node, end_node, dim, clockwise)
      last = nil
      if clockwise == signed_area(data, start_node, end_node, dim) > 0
        ((start_node...end_node).each_slice(dim).collect { |i| i[0] }).each { |i|
          last = insert_node(i, data[i], data[i + 1], last)
        }
      else
        ((start_node...end_node).each_slice(dim).collect { |i| i[0] }).reverse.each { |i|
          last = insert_node(i, data[i], data[i + 1], last)
        }
      end
      if last && (last == last.next)
        remove_node(last)
        last = last.next
      end
      last
    end

    # Eliminate collinear or duplicate points.
    #
    # @param start_node Add description
    # @param end_node Add description
    #
    # @example
    #   Add example.
    def self.filter_points(start_node, end_node = nil)
      return start_node unless start_node
      end_node = start_node unless end_node
      p = start_node
      again = true
      while again || (p != end_node)
        again = false
        if !p.steiner && equals(p, p.next) || area(p.prev, p, p.next) == 0
          remove_node(p)
          p = end_node = p.prev
          if p == p.next
            return nil
          end
          again = true
        else
          p = p.next
        end
      end
      end_node
    end

    # Main ear slicing loop which triangulates a polygon (given as a linked _list)
    #
    # @param ear Add description
    # @param triangles Add description
    # @param dim Add description
    # @param min_x Add description
    # @param min_y Add description
    # @param size Add description
    # @param _pass Add description
    #
    # @example
    #   Add example.
    def self.earcut_linked(ear, triangles, dim, min_x, min_y, size, _pass = nil)
      unless ear
        return
      end
      # interlink polygon nodes in z-order
      if !_pass && size
        index_curve(ear, min_x, min_y, size)
      end
      stop = ear
      prev = nil
      nxt = nil
      # iterate through ears, slicing them one by one
      count = 0
      while ear.prev != ear.next
        count += 1
        prev = ear.prev
        nxt = ear.next
        if size ? is_ear_hashed(ear, min_x, min_y, size) : is_ear(ear)
          # if isEarHashed(ear, minX, minY, size) if size else isEar(ear):
          # cut off the triangle
          triangles.push(prev.i / dim)
          triangles.push(ear.i / dim)
          triangles.push(nxt.i / dim)
          remove_node(ear)
          # skipping the next vertices leads to less sliver triangles
          ear = nxt.next
          stop = nxt.next
          next
        end
        ear = nxt
        # if we looped through the whole remaining polygon and can't find any more ears
        if ear == stop
          # try filtering points and slicing again
          if !_pass
            earcut_linked(filter_points(ear), triangles, dim, min_x, min_y, size, 1)
          else
            # if this didn't work, try curing all small self-intersections locally
            if _pass == 1
              ear = cure_local_intersections(ear, triangles, dim)
              earcut_linked(ear, triangles, dim, min_x, min_y, size, 2)
            else
              # as a last resort, try splitting the remaining polygon into two
              if _pass == 2
                split_earcut(ear, triangles, dim, min_x, min_y, size)
              end
            end
          end
          break
        end
      end
    end

    # Check whether a polygon node forms a valid ear with adjacent nodes.
    #
    # @param ear Add description
    #
    # @example
    #   Add example.
    def self.is_ear(ear)
      a = ear.prev
      b = ear
      c = ear.next
      if area(a, b, c) >= 0
        # reflex, can't be an ear
        return false
      end
      # now make sure we don't have other points inside the potential ear
      p = ear.next.next
      while p != ear.prev
        if point_in_triangle(a.x, a.y, b.x, b.y, c.x, c.y, p.x, p.y) && area(p.prev, p, p.next) >= 0
          return false
        end
        p = p.next
      end
      true
    end

    #  Add description.
    #
    # @param ear Add description
    # @param min_x Add description
    # @param min_y Add description
    # @param size Add description
    #
    # @example
    #   Add example.
    def self.is_ear_hashed(ear, min_x, min_y, size)
      a = ear.prev
      b = ear
      c = ear.next
      if area(a, b, c) >= 0
        # reflex, can't be an ear
        return false
      end
      # triangle bbox; min & max are calculated like this for speed
      min_tx = if a.x < b.x
                 (a.x < c.x) ? a.x : c.x
               else
                 (b.x < c.x) ? b.x : c.x
               end
      min_ty = if a.y < b.y
                 (a.y < c.y) ? a.y : c.y
               else
                 (b.y < c.y) ? b.y : c.y
               end
      max_tx = if a.x > b.x
                 (a.x > c.x) ? a.x : c.x
               else
                 (b.x > c.x) ? b.x : c.x
               end
      max_ty = if a.y > b.y
                 (a.y > c.y) ? a.y : c.y
               else
                 (b.y > c.y) ? b.y : c.y
               end
      # z-order range for the current triangle bbox;
      min_z = z_order(min_tx, min_ty, min_x, min_y, size)
      max_z = z_order(max_tx, max_ty, min_x, min_y, size)
      # first look for points inside the triangle in increasing z-order
      p = ear.next_z
      while p && p.z <= max_z
        if p != ear.prev && p != ear.next && point_in_triangle(a.x, a.y, b.x, b.y, c.x, c.y, p.x, p.y) && area(p.prev, p, p.next) >= 0
          return false
        end
        p = p.next_z
      end
      # then look for points in decreasing z-order
      p = ear.prev_z
      while p && p.z >= min_z
        if p != ear.prev && p != ear.next && point_in_triangle(a.x, a.y, b.x, b.y, c.x, c.y, p.x, p.y) && area(p.prev, p, p.next) >= 0
          return false
        end
        p = p.prev_z
      end
      true
    end

    # Go through all polygon nodes and cure small local self-intersections.
    def self.cure_local_intersections(start_node, triangles, dim)
      cure = true
      p = start_node
      while cure || p != start_node
        cure = false
        a = p.prev
        b = p.next.next
        if !equals(a, b) && intersects(a, p, p.next, b) && locally_inside(a, b) && locally_inside(b, a)
          triangles.push(a.i / dim)
          triangles.push(p.i / dim)
          triangles.push(b.i / dim)
          # remove two nodes involved
          remove_node(p)
          remove_node(p.next)
          p = start_node = b
        end
        p = p.next
      end
      p
    end

    # Try splitting polygon into two and triangulate them independently.
    def self.split_earcut(start_node, triangles, dim, min_x, min_y, size)
      # look for a valid diagonal that divides the polygon into two
      split = true
      a = start_node
      while split || a != start_node
        split = false
        b = a.next.next
        while b != a.prev
          if a.i != b.i && is_valid_diagonal(a, b)
            # split the polygon in two by the diagonal
            c = split_polygon(a, b)
            # filter collinear points around the cuts
            a = filter_points(a, a.next)
            c = filter_points(c, c.next)
            # run earcut on each half
            earcut_linked(a, triangles, dim, min_x, min_y, size)
            earcut_linked(c, triangles, dim, min_x, min_y, size)
            return
          end
          b = b.next
        end
        a = a.next
      end
    end

    # link every hole into the outer loop, producing a single-ring polygon
    # without holes.
    def self.eliminate_holes(data, hole_indicies, outer_node, dim)
      # puts "eliminateHoles called"
      queue = []
      _len = hole_indicies.size
      start_node = nil
      end_node = nil
      _list = nil
      hole_indicies.each_with_index do |hole, i|
        start_node = hole_indicies[i] * dim
        end_node = (i < _len - 1) ? (hole_indicies[i + 1]) * dim : data.size
        _list = linked_list(data, start_node, end_node, dim, false)
        if _list == _list.next
          _list.steiner = true
        end
        queue.push(get_leftmost(_list))
      end
      queue = queue.sort_by {|obj| obj.x}
      # process holes from left to right
      queue.each_with_index do |hole, i|
        eliminate_hole(queue[i], outer_node)
        outer_node = filter_points(outer_node, outer_node.next)
      end
      outer_node
    end

    # Find a bridge between vertices that connects hole with an outer ring and
    # and link it.
    def self.eliminate_hole(hole, outer_node)
      outer_node = find_hole_bridge(hole, outer_node)
      if outer_node
        b = split_polygon(outer_node, hole)
        filter_points(b, b.next)
      end
    end

    # David Eberly's algorithm for finding a bridge between hole and outer polygon.
    def self.find_hole_bridge(hole, outer_node)
      find = true
      p = outer_node
      hx = hole.x
      hy = hole.y
      # negative infinity
      qx = -1.0 / 0.0
      m = nil

      # find a segment intersected by a ray from the hole's leftmost point to the left;
      # segment's endpoint with lesser x will be potential connection point
      while find || p != outer_node
        find = false
        if hy <= p.y && hy >= p.next.y && (p.next.y - p.y != 0)
          x = p.x + (((hy - p.y) * (p.next.x - p.x)) / (p.next.y - p.y).to_f)
          if x <= hx && x > qx
            qx = x
            if x == hx
              if hy == p.y
                return p
              end
              if hy == p.next.y
                return p.next
              end
            end
            m = (p.x < p.next.x) ? p : p.next
          end
        end
        p = p.next
      end
      unless m
        return nil
      end
      if hx == qx
        # hole touches outer segment; pick lower endpoint
        return m.prev
      end
      # look for points inside the triangle of hole point, segment intersection and endpoint;
      # if there are no points found, we have a valid connection;
      # otherwise choose the point of the minimum angle with the ray as connection point
      stop = m
      mx = m.x
      my = m.y
      # positive infinity
      tan_min = 1.0 / 0.0
      tan = nil
      p = m.next
      while p != stop
        hx_or_qx = (hy < my) ? hx : qx
        qx_or_hx = (hy < my) ? qx : hx
        if hx >= p.x && p.x >= mx && point_in_triangle(hx_or_qx, hy, mx, my, qx_or_hx, hy, p.x, p.y)
          tan = ((hy - p.y).abs) / (hx - p.x).to_f
          if tan < tan_min || tan == tan_min && p.x > m.x && locally_inside(p, hole)
            m = p
            tan_min = tan
          end
        end
        p = p.next
      end
      m
    end

    # Interlink polygon nodes in z-order.
    def self.index_curve(start_node, min_x, min_y, size)
      index = true
      p = start_node
      while index || p != start_node
        index = false
        if p.z == nil
          p.z = z_order(p.x, p.y, min_x, min_y, size)
        end
        p.prev_z = p.prev
        p.next_z = p.next
        p = p.next
      end
      p.prev_z.next_z = nil
      p.prev_z = nil
      sort_linked(p)
    end

    # Simon Tatham's linked _list merge sort algorithm
    # @see [http:#www.chiark.greenend.org.uk/~sgtatham/algorithms/_listsort.html]
    def self.sort_linked(_list)
      sort = true
      i = nil
      p = nil
      q = nil
      e = nil
      tail = nil
      num_merges = nil
      p_size = nil
      q_size = nil
      in_size = 1
      while sort || num_merges > 1
        sort = false
        p = _list
        _list = nil
        tail = nil
        num_merges = 0
        while p
          num_merges += 1
          q = p
          p_size = 0
          in_size.times.each { |_|
            p_size += 1
            q = q.next_z
            unless q
              break
            end
          }
          q_size = in_size
          while p_size > 0 || q_size > 0 && q
            if p_size == 0
              e = q
              q = q.next_z
              q_size -= 1
            else
              if q_size == 0 || !q
                e = p
                p = p.next_z
                p_size -= 1
              else
                if p.z <= q.z
                  e = p
                  p = p.next_z
                  p_size -= 1
                else
                  e = q
                  q = q.next_z
                  q_size -= 1
                end
              end
            end
            if tail
              tail.next_z = e
            else
              _list = e
            end
            e.prev_z = tail
            tail = e
          end
          p = q
        end
        tail.next_z = nil
        in_size *= 2
      end
      _list
    end

    # z-order of a point given coords and size of the data bounding box
    def self.z_order(x, y, min_x, min_y, size)
      # coords are transformed into non-negative 15-bit integer range
      x = ((32767 * (x - min_x)) * size).to_i
      y = ((32767 * (y - min_y)) * size).to_i
      # puts "x = #{x}"
      x = x | x << 8 & 0x00FF00FF
      x = x | x << 4 & 0x0F0F0F0F
      x = x | x << 2 & 0x33333333
      x = x | x << 1 & 0x55555555
      y = y | y << 8 & 0x00FF00FF
      y = y | y << 4 & 0x0F0F0F0F
      y = y | y << 2 & 0x33333333
      y = y | y << 1 & 0x55555555
      x | y << 1
    end

    # Find the leftmost node of a polygon ring.
    def self.get_leftmost(start_node)
      get = true
      p = start_node
      leftmost = start_node
      while get || p != start_node
        get = false
        if p.x < leftmost.x
          leftmost = p
        end
        p = p.next
      end
      leftmost
    end

    # Check if a point lies within a convex triangle.
    def self.point_in_triangle(ax, ay, bx, by, cx, cy, px, py)
      (((cx - px) * (ay - py)) - ((ax - px) * (cy - py)) >= 0) && (((ax - px) * (by - py)) - ((bx - px) * (ay - py)) >= 0) && (((bx - px) * (cy - py)) - ((cx - px) * (by - py)) >= 0)
    end

    # Check if a diagonal between two polygon nodes is valid (lies in polygon
    # interior).
    def self.is_valid_diagonal(a, b)
      a.next.i != b.i && a.prev.i != b.i && !intersects_polygon(a, b) && locally_inside(a, b) && locally_inside(b, a) && middle_inside(a, b)
    end

    # Signed area of a triangle.
    def self.area(p, q, r)
      ((q.y - p.y) * (r.x - q.x)) - ((q.x - p.x) * (r.y - q.y))
    end

    # Check if two points are equal.
    def self.equals(p1, p2)
      p1.x == p2.x && p1.y == p2.y
    end

    # Check if two segments intersect.
    def self.intersects(p1, q1, p2, q2)
      if equals(p1, q1) && equals(p2, q2) || equals(p1, q2) && equals(p2, q1)
        return true
      end
      (((area(p1, q1, p2) > 0) != (area(p1, q1, q2) > 0)) and
        ((area(p2, q2, p1) > 0) != (area(p2, q2, q1) > 0)))
    end

    # Check if a polygon diagonal intersects any polygon segments.
    def self.intersects_polygon(a, b)
      intersects = true
      p = a
      while intersects || p != a
        intersects = false
        if p.i != a.i && p.next.i != a.i && p.i != b.i && p.next.i != b.i && intersects(p, p.next, a, b)
          return true
        end
        p = p.next
      end
      false
    end

    # Check if a polygon diagonal is locally inside the polygon.
    def self.locally_inside(a, b)
      if area(a.prev, a, a.next) < 0
        return area(a, b, a.next) >= 0 && area(a, a.prev, b) >= 0
      else
        return area(a, b, a.prev) < 0 || area(a, a.next, b) < 0
      end
    end

    # Check if the middle point of a polygon diagonal is inside the polygon.
    def self.middle_inside(a, b)
      middle = true
      p = a
      inside = false
      px = (a.x + b.x) / 2.to_f
      py = (a.y + b.y) / 2.to_f
      while middle || p != a
        middle = false
        if p.y > py != p.next.y > py && (px < (((p.next.x - p.x) * (py - p.y)) / (p.next.y - p.y).to_f) + p.x)
          inside = !inside
        end
        p = p.next
      end
      inside
    end

    # Link two polygon vertices with a bridge; if the vertices belong to the
    # same ring, it splits polygon into two; if one belongs to the outer ring
    # and another to a hole, it merges it into a single ring.
    def self.split_polygon(a, b)
      a2 = Node.new(a.i, a.x, a.y)
      b2 = Node.new(b.i, b.x, b.y)
      an = a.next
      bp = b.prev
      a.next = b
      b.prev = a
      a2.next = an
      an.prev = a2
      b2.next = a2
      a2.prev = b2
      bp.next = b2
      b2.prev = bp
      b2
    end

    # Create a node and optionally link it with previous one (in a circular
    # doubly linked _list).
    def self.insert_node(i, x, y, last)
      p = Node.new(i, x, y)
      if !last
        p.prev = p
        p.next = p
      else
        p.next = last.next
        p.prev = last
        last.next.prev = p
        last.next = p
      end
      p
    end

    def self.remove_node(p)
      p.next.prev = p.prev
      p.prev.next = p.next
      p.prev_z.next_z = p.next_z if p.prev_z
      p.next_z.prev_z = p.prev_z if p.next_z
    end

    # return a percentage difference between the polygon area and its
    # triangulation area; used to verify correctness of triangulation
    def self.deviation(data, hole_indicies, dim, triangles)
      _len = hole_indicies.size
      has_holes = hole_indicies[0] != nil
      outer_len = has_holes ? (hole_indicies[0] * dim) : data.length
      polygon_area = signed_area(data, 0, outer_len, dim).abs
      if has_holes
        _len.times.each { |i|
          start_node = hole_indicies[i] * dim
          end_node = (i < _len - 1) ? (hole_indicies[i + 1]) * dim : data.length
          polygon_area -= signed_area(data, start_node, end_node, dim).abs
        }
      end

      triangles_area = 0
      i = 0
      (triangles.length / 3).times do
        # for i in 0.step(triangles.size, 3)
        a = triangles[i] * dim
        b = (triangles[i + 1]) * dim
        c = (triangles[i + 2]) * dim
        triangles_area += (((data[a] - data[c]) * ((data[b + 1]) - (data[a + 1]))) - ((data[a] - data[b]) * ((data[c + 1]) - (data[a + 1])))).abs
        i += 3
      end

      if polygon_area == 0 && triangles_area == 0
        return 0
      end
      ((triangles_area - polygon_area) / polygon_area.to_f).abs
    end

    def self.signed_area(data, start_node, end_node, dim)
      sum = 0
      j = end_node - dim
      ((start_node...end_node).each_slice(dim).collect { |i| i[0] }).each { |i|
        next if i > end_node
        sum += (data[j] - data[i]) * ((data[i + 1]) + (data[j + 1]))
        j = i
      }
      sum
    end

    def self.unflatten(data)
      result = []
      0.step(data.size, 3).each { |i|
        result.push(data[i...i + 3].to_a)
      }
      result
    end
  end
end
